#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Utility for generating API objects files for tantrum.

For each directory in this scripts path that has a file console.wsdl:
    treat directory name as API version
    parse console.wsdl for API objects
    for each filename in [soap.py]:
        the source contents are combined with the parsed objects
        file is written to ${filename}_${api_version}.py in the parent directory
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import argparse
import black
import datetime
import importlib
import logging
import platform
import string
import sys
import xmltodict

logging.basicConfig(level=logging.INFO, format="%(levelname)s: %(message)s")
LOG = logging.getLogger()

try:
    import pathlib
except Exception:
    import pathlib2 as pathlib

__version__ = "3.0.0"

TMPL_BUILD_META = '''
BUILD_META = {
    "script": "${script}",
    # Script used to build this file
    "script_version": "${script_version}",
    # Version of script used to build this file
    "script_platform": "${script_platform}",
    # OS Platform that ran this build script
    "script_python": "${script_python}",
    # Python version that ran this build script
    "date": "${date}",
    # Date/time in UTC format of when this file was built
    "source_file": "${source_file}",
    # File that was used to auto-generate objects in this file
    "source_file_date": "${source_file_date}",
    # Date/time in UTC format of the source_file
}
""":obj:`dict`: How this module was built (date/time in UTC format)."""'''
TMPL_BUILD_META = string.Template(TMPL_BUILD_META)
""":obj:`string.Template`: Template used to build BUILD_META dict part."""

TMPL_VERSION = '''
VERSION = {
    "major": ${major},
    "minor": ${minor},
    "protocol": ${protocol},
    "build": ${build},
    "string": "${string}",
}
""":obj:`dict`: Tanium API version these objects are intended to be used for."""

__version__ = VERSION["string"]
""":obj:`str`: Tanium API version these objects are intended to be used for."""
'''
TMPL_VERSION = string.Template(TMPL_VERSION)
""":obj:`string.Template`: Template used to build VERSION dict part."""

TMPL_COMMANDS = '''
COMMANDS = [${commands}
]
""":obj:`list` of :obj:`str`:API commands."""'''
TMPL_COMMANDS = string.Template(TMPL_COMMANDS)
""":obj:`string.Template`: Template used to build COMMANDS list part."""

TMPL_LIST_CLS = '''
class ${class_name}(ApiList):
    """Automagically generated API array object."""

    API_NAME = "${api_name}"
    """:obj:`str`: Name of object used in API calls."""

    API_NAME_SRC = "${src_type}"
    """:obj:`str`: Name of object in source file."""

    API_SIMPLE = {${simple_parts}}
    """:obj:`dict`: Map of simple attributes to their types."""

    API_COMPLEX = {${complex_parts}}
    """:obj:`dict`: Map of complex attributes to their types."""

    API_CONSTANTS = {${constant_parts}}
    """:obj:`dict`: Map of constants to their values."""

    API_STR_ADD = []
    """:obj:`list` or :obj:`str`: Attributes to add to str formatting."""

    API_ITEM_ATTR = "${item_attr}"
    """:obj:`str`: Name of :attr:`API_ITEM_CLS` used in API calls."""

    API_ITEM_CLS = "${item_cls}"
    """:class:`ApiItem`: Item class this list class holds."""
'''
TMPL_LIST_CLS = string.Template(TMPL_LIST_CLS)
""":obj:`string.Template`: Template used to build ApiList class objects."""

TMPL_ITEM_CLS = '''
class ${class_name}(ApiItem):
    """Automagically generated API object."""

    API_NAME = "${api_name}"
    """:obj:`str`: Name of object used in API calls."""

    API_NAME_SRC = "${src_type}"
    """:obj:`str`: Name of object in source file."""

    API_SIMPLE = {${simple_parts}}
    """:obj:`dict`: Map of simple attributes to their types."""

    API_COMPLEX = {${complex_parts}}
    """:obj:`dict`: Map of complex attributes to their types."""

    API_CONSTANTS = {${constant_parts}}
    """:obj:`dict`: Map of constants to their values."""

    API_STR_ADD = []
    """:obj:`list` or :obj:`str`: Attributes to add to str formatting."""

    API_LIST_CLS = "${list_cls}"
    """:class:`ApiList`: List class that holds this item class."""

    API_LIST_API_NAME = "${list_name}"
    """:obj:`str`: Name of :attr:`API_LIST_CLS` used in API calls."""
'''
TMPL_ITEM_CLS = string.Template(TMPL_ITEM_CLS)
""":obj:`string.Template`: Template used to build ApiItem class objects."""

TMPL_ATTR = """        "{name}": "{type_camel}",  # noqa: E501"""
""":obj:`str`: String formatter used to build simple/complex attrs."""

TMPL_CONSTANT = """        "{name}": {value!r},"""
""":obj:`str`: String formatter used to build constants."""

TMPL_MODULE = """${input_part}
${parsed_part}
api_fixes()
expand_cls_globals()
"""
TMPL_MODULE = string.Template(TMPL_MODULE)
""":obj:`string.Template`: Template used to build the output file."""

TMPL_DOC = """{TYPE} API Object Module for Version {major}.{minor}.{protocol}.{build}
#####################################################################################

.. automodule:: tantrum.api_objects.{type}_{major}_{minor}_{protocol}_{build}
    :members:
    :undoc-members:
    :show-inheritance:
    :private-members:
    :member-order: bysource
"""
""":obj:`str`: String formatter used to build sphinx doc file."""

TMPL_VER_STR = "{major}_{minor}_{protocol}_{build}"
""":obj:`str`: String formatter used to build version str for file name."""

TMPL_FILE_STEM = "{api_type}_{ver_str}"
""":obj:`str`: String formatter used to build stem str for file name."""


def get_paths_with(path, with_file, only=None):
    """Get all directories under path that have with_file in them."""
    path = pathlib.Path(path).absolute()
    paths = [p for p in path.glob("*") if p.is_dir() and (p / with_file).is_file()]
    if only:
        paths = [p for p in paths if p.name == only]
    if not paths:
        if only:
            error = "No paths exist that only match '{o}' with file '{f}' under '{p}'"
        else:
            error = "No paths exist with file '{f}' under '{p}'"
        error = error.format(f=with_file, p=path / "*", o=only)
        raise Exception(error)
    return paths


def write_path(path, contents):
    """Write the contents of a file.."""
    path.write_text(contents)
    m = "Wrote {c} bytes to '{p}'"
    m = m.format(c=len(contents), p=path)
    LOG.info(m)


def get_input_path_contents(path):
    """Get the contents of a file."""
    if not path.is_file():
        error = "Unable to find file: '{p}'"
        error = error.format(p=path)
        raise Exception(error)
    contents = path.read_text()
    return contents


def parse_version_from_path(path):
    """Get version parts from a path name."""
    path = pathlib.Path(path).absolute()
    version = path.name
    try:
        parts = version.split("_")
        ret = {}
        ret["major"] = try_int(parts[0])
        ret["minor"] = try_int(parts[1])
        ret["protocol"] = try_int(parts[2])
        ret["build"] = try_int(parts[3])
        ret["string"] = version.replace("_", ".")
        ret["file"] = version
    except Exception:
        error = "Bad API version in '{p}', must look like: '7_2_314_3181'"
        error = error.format(p=path)
        raise Exception(error)
    return ret


def camelcase(text):
    """Convert some_string or some-string to SomeString."""
    text_spaced = text.replace("_", " ").replace("-", " ")
    ret = string.capwords(text_spaced).replace(" ", "")
    return ret


def try_int(value):
    """Coerce an object into an int."""
    try:
        return int(value)
    except Exception:
        return value


def coerce_list(o):
    """Coerce an object into a list."""
    if not isinstance(o, (list, tuple)):
        o = [o]
    return o


def add_args(argparser, args):
    """Add arguments to an argparser based off of list of dicts in args."""
    required = argparser.add_argument_group("Required Arguments")
    optional = argparser.add_argument_group("Optional Arguments")

    for arg in args:
        arg_required = arg.get("required", False)
        arg_default = arg.get("default", argparse.SUPPRESS)
        arg_help = arg.get("help", "")
        arg_opts = arg.pop("opts") if "opts" in arg else []

        grp = required if arg_required else optional

        helps = []
        helps.append(arg_help)

        if arg_default is not argparse.SUPPRESS and not arg_required:
            helps.append("DEFAULT: '{d}'".format(d=arg_default))

        choices = arg.get("choices", [])
        if choices:
            helps.append("VALID VALUES: {c}".format(c=", ".join(choices)))

        arg["help"] = "\n".join(helps)
        grp.add_argument(*arg_opts, **arg)
    return argparser


def arg_parse(args):
    """Parse arguments from command line."""
    argparser = argparse.ArgumentParser(
        description=__doc__,
        add_help=False,
        formatter_class=argparse.RawTextHelpFormatter,
    )
    add_args(argparser=argparser, args=args)
    args, extra_args = argparser.parse_known_args()
    return args, extra_args


def check_dest_path(path, overwrite=False):
    """Check if path exists."""
    if path.is_file():
        error = "Overwrite is {o} and found pre-existing file: '{p}'"
        error = error.format(o=overwrite, p=path)
        if not overwrite:
            raise Exception(error)
        else:
            LOG.warning(error)


class WsdlParser(object):
    """Parse a WSDL file to find all API objects."""

    object_skips = ["TaniumSOAPResult", "TaniumSOAPRequest", "auth"]
    # objects to skip over when parsing objects

    typemap = {
        "xsd:string": "string_types",
        "xsd:int": "integer_types",
        "xsd:long": "integer_types",
        "xsd:double": "float_types",
    }
    # mapping of xsd types to python types

    simpletypes = ["string_types", "integer_types", "float_types"]

    override_names = {
        "effective_content_set_privilege_request": "effective_content_set_privilege_request"  # noqa
    }

    def __init__(self, text):
        """Constructor."""
        self.src_text = text

        try:
            self.root = xmltodict.parse(self.src_text)
        except Exception as e:
            m = "Unable to parse XML text, error: {e}".format(e)
            raise Exception(m)
        self.schema = self.root["definitions"]["types"]["xsd:schema"]
        olist = [
            x for x in self.schema["xsd:complexType"] if x["@name"] == "object_list"
        ][0]
        olist_els = olist["xsd:sequence"]["xsd:element"]
        self.objlistmap = {x["@type"]: x["@name"] for x in olist_els}

    def get_commands(self):
        """Parse the list of valid commands from the WSDL."""
        simple_types = self.schema["xsd:simpleType"]
        elems = simple_types["xsd:restriction"]["xsd:enumeration"]
        commands = [x["@value"] for x in elems]
        return commands

    def add_list_cls_to_items(self, objects):
        """Try to determine the list class that owns an item class."""
        list_objects = [x for x in objects if x["list_attr"]]
        complex_list_objects = [
            x for x in list_objects if not x["list_attr"]["is_simple"]
        ]
        item_objects = [x for x in objects if not x["list_attr"]]
        mismatch1 = []
        mismatch2 = []
        for obj in complex_list_objects:
            list_attr = obj["list_attr"]
            if list_attr["name"] in obj["api_name"]:
                item_obj = [
                    x for x in objects if x["class_name"] == list_attr["type_camel"]
                ]
                item_obj[0]["list_cls"] = obj["class_name"]
                item_obj[0]["list_name"] = obj["api_name"]
            else:
                mismatch1.append(obj)

        for obj in mismatch1:
            list_attr = obj["list_attr"]
            item_obj = [x for x in item_objects if x["src_type"] == list_attr["type"]]
            if item_obj:
                if "list_cls" not in item_obj[0]:
                    item_obj[0]["list_cls"] = obj["class_name"]
                    item_obj[0]["list_name"] = obj["api_name"]
            else:
                mismatch2.append(obj)

        for obj in item_objects:
            obj["list_cls"] = obj.get("list_cls", None)
            obj["list_name"] = obj.get("list_name", "")
        return objects

    def get_objects(self):
        """Parse the list of complex types into a list of dicts."""
        objects = []

        for ctype in self.schema["xsd:complexType"]:
            if ctype["@name"] in self.object_skips:
                continue
            obj = self._parse_complex_type(ctype)
            if not obj["class_name"] in [x["class_name"] for x in objects]:
                objects.append(obj)

        objects = self.add_list_cls_to_items(objects)
        return objects

    def _parse_complex_type(self, ctype):
        """Parse a complex type into a dict.

        Should produce a dictionary as follows:
            src_type: str
                object type as known in the wsdl
            api_name: str
                object name used in API requests/responses
            class_name: str
                object name of Python class
            constants: list of dict
                object constants defined in wsdl, if any
            simple_attrs: list of dict
                object attributes that are simple (str, int, float)
            complex_attrs: list of dict
                object attributes that are not simple (other objects)
            if object is a list object:
                list_attr: dict
                    attribute used to store items
                item_cls: str
                    camel type of items this list class stores
                item_attr: str
                    attribute name used in API requests/responses
            if object is not a list object:
                list_cls: str
                    camel type of list_cls that holds this item

        Each attribute dict in list_attr, simple_attrs, and complex_attrs is as follows:
            name: str
                name of attribute
            type: str
                type of object attribute stores (string/int/float/ or other API object)
            type_camel: str
                camel case type if type is complex, else just type
            is_list: bool
                this attribute is an array
            is_simple: bool
                this attribute is simple
        """
        obj = {}
        obj["src_type"] = ctype["@name"]
        obj["class_name"] = self._camel(ctype["@name"])
        obj["api_name"] = self._get_obj_name(ctype["@name"])
        obj["constants"] = []
        obj["list_attr"] = {}

        attrs = []
        xsd_all = ctype.get("xsd:all", {})
        xsd_seq = ctype.get("xsd:sequence", {})
        xsd_attr = ctype.get("xsd:attribute", [])

        xsd_all_els = coerce_list(xsd_all.get("xsd:element", []) or [])
        xsd_seq_els = coerce_list(xsd_seq.get("xsd:element", []) or [])
        xsd_attr_els = coerce_list(xsd_attr)

        all_els = xsd_all_els + xsd_seq_els + xsd_attr_els
        all_els = [x for x in all_els if x]

        for el in all_els:
            el_maxoccurs = el.get("@maxOccurs", "")
            is_unbounded = el_maxoccurs == "unbounded"
            is_list_ending = ctype["@name"].endswith("_list")
            is_one_seq = len(xsd_seq_els) == 1
            is_list = is_unbounded or (is_list_ending and is_one_seq)
            attr = {}
            attr["name"] = el["@name"]
            attr["type"] = self.typemap.get(el["@type"], el["@type"])
            attr["type_camel"] = self._camel(attr["type"])
            attr["is_list"] = is_list
            attr["is_simple"] = attr["type"] in self.simpletypes
            attrs.append(attr)

        xsd_cons = ctype.get("xsd:annotation", {})
        xsd_cons = xsd_cons["xsd:appinfo"]["constants"]["constant"] if xsd_cons else []

        obj["constants"] = [
            {"name": el["name"], "value": try_int(el["value"]["#text"])}
            for el in xsd_cons
        ]

        list_attrs = [x for x in attrs if x["is_list"]]
        if len(list_attrs) == 1:
            obj["list_attr"] = list_attrs[0]
            obj["item_cls"] = obj["list_attr"]["type_camel"]
            obj["item_attr"] = obj["list_attr"]["name"]
            attrs = [x for x in attrs if x is not list_attrs[0]]

        obj["simple_attrs"] = [x for x in attrs if x["is_simple"]]
        obj["complex_attrs"] = [x for x in attrs if not x["is_simple"]]
        return obj

    def _get_obj_name(self, name):
        """Fancy dance to figure out the name that is used in requests/responses."""
        if name in self.override_names:
            return self.override_names[name]
        if name in self.objlistmap:
            return self.objlistmap[name]
        if name.endswith("_list") and name != "object_list":
            return name.replace("_list", "") + ("es" if name.endswith("s") else "s")
        return name

    def _camel(self, t):
        return t if t in self.simpletypes else camelcase(t)


def build_class_parts(objects, tmpl):
    """Build a class object."""
    parts = []
    for obj in objects:
        obj["constant_parts"] = build_cls_attr(
            items=obj["constants"], tmpl=TMPL_CONSTANT
        )
        obj["simple_parts"] = build_cls_attr(items=obj["simple_attrs"], tmpl=TMPL_ATTR)
        obj["complex_parts"] = build_cls_attr(
            items=obj["complex_attrs"], tmpl=TMPL_ATTR
        )
        part = tmpl.substitute(**obj)
        parts.append(part)
    return "\n".join(parts)


def build_meta_part(name, path, src_url):
    """Build the BUILD_META dict."""
    path = pathlib.Path(path)
    path_stat = pathlib.Path(path).stat()
    path_mtime = datetime.datetime.utcfromtimestamp(path_stat.st_mtime)
    now = datetime.datetime.utcnow()
    meta = {}
    meta["script"] = name
    meta["script_version"] = __version__
    meta["script_platform"] = platform.platform()
    meta["script_python"] = sys.version.splitlines()[0]
    meta["date"] = "{}".format(now)
    meta["source_file"] = "{}".format(src_url)
    meta["source_file_date"] = "{}".format(path_mtime)
    return TMPL_BUILD_META.substitute(**meta)


def build_version_part(version):
    """Build VERSION dict."""
    return TMPL_VERSION.substitute(**version)


def build_commands_part(commands):
    """Build COMMANDS list."""
    cmds = ["    {!r},".format(c) for c in commands]
    cmds = ([""] + cmds) if cmds else cmds
    return TMPL_COMMANDS.substitute(commands="\n".join(cmds))


def build_cls_attr(items, tmpl):
    """Build class attributes."""
    ret = [tmpl.format(**i) for i in items]
    if ret:
        ret = [""] + ret + ["    "]
    return "\n".join(ret)


def build_parts(path, file, version):
    """Generate all of the parts."""
    module = importlib.import_module(version["file"])
    wsdl_parser = WsdlParser(text=module.data)

    commands = wsdl_parser.get_commands()
    all_objects = wsdl_parser.get_objects()
    list_objects = [x for x in all_objects if x["list_attr"]]
    item_objects = [x for x in all_objects if not x["list_attr"]]

    m = "Parsed {i} item objects, {co} list objects, and {cc} commands from '{p}'"
    m = m.format(i=len(item_objects), co=len(list_objects), cc=len(commands), p=path)
    LOG.debug(m)

    src_url = module.from_url_path
    parts = [
        build_class_parts(objects=item_objects, tmpl=TMPL_ITEM_CLS),
        build_class_parts(objects=list_objects, tmpl=TMPL_LIST_CLS),
        build_meta_part(name=this_script.name, path=path, src_url=src_url),
        build_commands_part(commands=commands),
        build_version_part(version=version),
    ]
    return "\n".join(parts)


if __name__ == "__main__":
    this_script = pathlib.Path(sys.argv[0]).expanduser().absolute()
    script_args = [
        {
            "opts": ["-o", "--overwrite"],
            "dest": "overwrite",
            "default": False,
            "action": "store_true",
            "help": "Overwrite output objects files that already exist.",
            "required": False,
        },
        {
            "opts": ["-b", "--build"],
            "dest": "build_dir",
            "default": None,
            "help": "Only build for this version instead of for all.",
            "required": False,
        },
        {
            "opts": ["-d", "--debug"],
            "dest": "debug",
            "default": False,
            "action": "store_true",
            "help": "Show debug logging.",
            "required": False,
        },
        {
            "opts": ["--version"],
            "help": "Version of this script.",
            "version": "%(prog)s {v}".format(v=__version__),
            "default": argparse.SUPPRESS,
            "action": "version",
        },
        {
            "opts": ["-h", "--help"],
            "help": "Show this help message and exit.",
            "default": argparse.SUPPRESS,
            "action": "help",
        },
    ]
    args, _ = arg_parse(args=script_args)

    LOG.setLevel(logging.DEBUG if args.debug else logging.INFO)

    build_file = "__init__.py"
    get_path = this_script.parent
    paths = get_paths_with(path=get_path, with_file=build_file, only=args.build_dir)

    obj_files = ["soap.py"]
    output_path = this_script.parent.parent
    pkg_root = output_path.parent.parent
    docs_output_path = pkg_root / "docs" / "api" / "api_objects"

    for path in paths:
        version = parse_version_from_path(path=path)
        version_str = TMPL_VER_STR.format(**version)
        parsed_part = build_parts(path=path, file=build_file, version=version)
        for obj_file in obj_files:
            input_path = path / obj_file
            try:
                input_part = get_input_path_contents(path=input_path)
            except Exception as exc:
                error = "Build error, skipping build for '{p}' - error:\n!!  {e}"
                error = error.format(p=input_path, e=exc)
                LOG.error(error)
                continue

            api_type = input_path.stem
            file_stem = TMPL_FILE_STEM.format(api_type=api_type, ver_str=version_str)

            py_contents = TMPL_MODULE.substitute(
                input_part=input_part, parsed_part=parsed_part
            )
            py_contents = py_contents.strip()
            mode = black.FileMode(target_versions=black.PY36_VERSIONS)
            py_contents = black.format_file_contents(
                src_contents=py_contents, fast=False, mode=mode
            )

            rst_dest_file = "{stem}.rst".format(stem=file_stem)
            rst_dest_path = docs_output_path / api_type / rst_dest_file
            rst_contents = TMPL_DOC.format(
                TYPE=api_type.upper(), type=api_type, **version
            )
            write_path(contents=rst_contents, path=rst_dest_path)

            py_dest_file = "{stem}.py".format(stem=file_stem)
            py_dest_path = output_path / py_dest_file
            try:
                check_dest_path(path=py_dest_path, overwrite=args.overwrite)
            except Exception as exc:
                error = "Skipping build for '{p}':\n!!  {e}"
                error = error.format(p=input_path, e=exc)
                LOG.error(error)
                continue
            write_path(contents=py_contents, path=py_dest_path)



# -*- coding: utf-8 -*-
"""Python objects for Tanium's API."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import datetime
import importlib
import six
import warnings

from . import exceptions
from .. import api_models
from .. import utils

if six.PY2:
    import pathlib2 as pathlib  # pragma: no cover
else:
    import pathlib

API_TYPES = ["soap"]
""":obj:`list` of :obj:`str`: All API object module types supported."""

PATTERN = "{api_type}_*.py"
""":obj:`str`: Glob pattern to find modules in :func:`get_versions`."""

DEFAULT_TYPE = "soap"
""":obj:`str`: Default :data:`API_TYPES` type to load in :func:`load`."""


class ApiObjects(object):
    """Encapsulation object for an API Object module."""

    def __init__(self, module_file, adhoc_warn=True):
        """Constructor.

        Args:
            module_file (:obj:`str`):
                API Object module file name in import string format.
            adhoc_warn (:obj:`bool`, optional):
                Enable warnings about adhoc added simple attributes to API Objects.

                Defaults to: True.

        """
        module_path = "{p}.{f}".format(p=__name__, f=module_file)
        self._module = importlib.import_module(module_path)
        """:mod:`tantrum.api_objects`: Imported API Objects module."""

        self._models = api_models
        """:mod:`tantrum.api_models`: API Models module."""

        for cls in self.cls_all:
            cls.API_OBJECTS = self
            setattr(self, cls.__name__, cls)

        self.control_adhoc_warnings(adhoc_warn)

    def __str__(self):
        """Show object info.

        Returns:
            :obj:`str`

        """
        bits = [
            "type={!r}".format(self.module_type),
            "version={!r}".format(self.module_version),
        ]
        bits = "({})".format(", ".join(bits))
        cls = "{c.__module__}.{c.__name__}".format(c=self.__class__)
        return "{cls}{bits}".format(cls=cls, bits=bits)

    def __repr__(self):
        """Show object info.

        Returns:
            :obj:`str`

        """
        return self.__str__()

    def control_adhoc_warnings(self, enable=False):
        """Enable or disable warnings about adhoc added simple attributes.

        Args:
            enable (:obj:`bool`, optional):
                Enable warnings.

                Defaults to: True.

        """
        action = "default" if enable else "ignore"
        warnings.simplefilter(action, api_models.exceptions.AttrUndefinedWarning)

    @property
    def ApiModel(self):
        """Get the ApiModel.

        Returns:
            :class:`tantrum.api_models.ApiModel`

        """
        return self.module.ApiModel

    @property
    def ApiItem(self):
        """Get the ApiItem.

        Returns:
            :class:`tantrum.api_models.ApiItem`

        """
        return self.module.ApiItem

    @property
    def ApiList(self):
        """Get the ApiList.

        Returns:
            :class:`tantrum.api_models.ApiList`

        """
        return self.module.ApiList

    @property
    def module(self):
        """Get the API objects module.

        Returns:
            :obj:`object`

        """
        return self._module

    @property
    def module_type(self):
        """Get the API type.

        Returns:
            :obj:`str`

        """
        return self.module.TYPE

    @property
    def module_version(self):
        """Get the API version.

        Returns:
            :obj:`str`

        """
        return self.module.__version__

    @property
    def module_dt(self):
        """Get the date time format string.

        Returns:
            :obj:`str`

        """
        return self.module.API_DT

    def module_dt_format(self, dt):
        """Format a datetime string using :attr:`module_dt`.

        Args:
            dt (:obj:`str`):
                String to format.

        Returns:
            :obj:`datetime.datetime`

        """
        return datetime.datetime.strptime(dt, self.module_dt)

    @property
    def module_version_dict(self):
        """Get the API version.

        Returns:
            :obj:`dict`

        """
        return self.module.VERSION

    @property
    def cls_item(self):
        """Get all of the ApiItem classes.

        Returns:
            :obj:`list` of :class:`tantrum.api_models.ApiItem`

        """
        return self.ApiItem.__subclasses__()

    @property
    def cls_list(self):
        """Get all of the ApiItem classes.

        Returns:
            :obj:`list` of :class:`tantrum.api_models.ApiList`

        """
        return self.ApiList.__subclasses__()

    @property
    def cls_all(self):
        """Get all of the ApiItem and ApiList classes.

        Returns:
            :obj:`list` of :class:`tantrum.api_models.ApiModel`

        """
        return self.cls_list + self.cls_item

    @property
    def cls_name_map_item(self):
        """Get a map of API key name to class for all ApiItem classes.

        Returns:
            :obj:`dict`

        """
        name_map = {}
        for c in self.cls_item:
            if c.API_NAME in name_map:
                other = name_map[c.API_NAME]
                error = "{cls} with API name {name!r} already mapped as {other}"
                error = error.format(cls=c, name=c.API_NAME, other=other)
                raise exceptions.ModuleError(error)
            name_map[c.API_NAME] = c
        return name_map

    @property
    def cls_name_map_list(self):
        """Get a map of API key name to class for all ApiList classes.

        Returns:
            :obj:`dict`

        """
        name_map = {}
        for c in self.cls_list:
            if c.API_NAME in name_map:
                other = name_map[c.API_NAME]
                error = "{cls} with API name {name!r} already mapped as {other}"
                error = error.format(cls=c, name=c.API_NAME, other=other)
                raise exceptions.ModuleError(error)
            name_map[c.API_NAME] = c
        return name_map

    @property
    def cls_name_map_all(self):
        """Get a map of API key name to class for all ApiItem and ApiList classes.

        Returns:
            :obj:`dict`

        """
        name_map = {}
        for c in self.cls_all:
            if c.API_NAME in name_map:
                other = name_map[c.API_NAME]
                error = "{cls} with API name {name!r} already mapped as {other}"
                error = error.format(cls=c, name=c.API_NAME, other=other)
                raise exceptions.ModuleError(error)
            name_map[c.API_NAME] = c
        return name_map

    def cls_item_by_name(self, name):
        """Get an ApiItem class by API name.

        Args:
            name (:obj:`str`):
                Name of object used in API calls.

        Returns:
            :class:`tantrum.api_models.ApiItem`

        """
        name_map = self.cls_name_map_item
        if name not in name_map:
            raise exceptions.UnknownApiNameError(
                name=name, name_map=name_map, module=self.module
            )
        return name_map[name]

    def cls_list_by_name(self, name):
        """Get an ApiList class by API name.

        Args:
            name (:obj:`str`):
                Name of object used in API calls.

        Returns:
            :class:`tantrum.api_models.ApiList`

        """
        name_map = self.cls_name_map_list
        if name not in name_map:
            raise exceptions.UnknownApiNameError(
                name=name, name_map=name_map, module=self.module
            )
        return name_map[name]

    def cls_by_name(self, name):
        """Get an ApiItem or ApiList by API name.

        Args:
            name (:obj:`str`):
                Name of object used in API calls.

        Returns:
            :class:`tantrum.api_models.ApiModel`

        """
        name_map = self.cls_name_map_all
        if name not in name_map:
            raise exceptions.UnknownApiNameError(
                name=name, name_map=name_map, module=self.module
            )
        return name_map[name]

    @classmethod
    def load(cls, api_type=DEFAULT_TYPE, veq="", vmax="", vmin="", vshrink=True):
        """Import an API module for version and type.

        Args:
            api_type (:obj:`str`, optional):
                Type of object module to load, must be one of :data:`API_TYPES`.

                Defaults to: :data:`DEFAULT_TYPE`.
            veq (:obj:`str`, optional):
                Exact version that API object module file must match.

                Defaults to: "".
            vmin (:obj:`str`, optional):
                Minimum version that API object module file must match.

                Defaults to: "".
            vmax (:obj:`str`, optional):
                Maximum version that API object module file must match.

                Defaults to: "".
            vshrink (:obj:`bool` or :obj:`int`, optional):
                If True, shrink API object module file version down to match the
                same length as veq, vmax, and vmin.

                If False, do not shrink API object module file version at all.

                If int, shrink API object module file version, veq, vmax, and vmin
                down to this length.

                Defaults to: True.

        Raises:
            :exc:`exceptions.ModuleError`:
                If api_type not in :data:`API_TYPES`.

        Returns:
            :obj:`ApiObjects`

        """
        return load(api_type=api_type, veq=veq, vmax=vmax, vmin=vmin, vshrink=vshrink)


def get_versions(api_type=DEFAULT_TYPE):
    """Search for API object module files of api_type.

    Args:
        api_type (:obj:`str`, optional):
            Type of object module to load, must be one of :data:`API_TYPES`.

            Defaults to: :data:`DEFAULT_TYPE`.

    Raises:
        :exc:`exceptions.NoVersionFoundError`:
              If no API module files matching :data:`PATTERN` are found.

    Returns:
        :obj:`list` of :obj:`dict`

    """
    path = pathlib.Path(__file__).absolute().parent
    pattern = PATTERN.format(api_type=api_type)
    matches = [p for p in path.glob(pattern)]

    if not matches:
        error = "Unable to find any object modules matching pattern {r!r} in {p!r}"
        error = error.format(p=format(path), r=pattern)
        raise exceptions.NoVersionFoundError(error)

    versions = []
    for match in matches:
        name = match.stem
        vparts = name.split("_")
        vtype = vparts.pop(0)
        vparts = utils.versions.split_ver(vparts)
        vstr = utils.versions.join_ver(vparts)
        versions.append(
            {
                "ver_str": vstr,
                "ver_parts": vparts,
                "api_type": vtype,
                "module_file": name,
                "module_path": match,
            }
        )

    versions = sorted(versions, key=lambda x: x["ver_parts"], reverse=True)
    return versions


def find_version(api_type=DEFAULT_TYPE, veq="", vmax="", vmin="", vshrink=True):
    """Find a API module files that match version and type.

    Args:
        api_type (:obj:`str`, optional):
            Type of object module to find, must be one of :data:`API_TYPES`.

            Defaults to: :data:`DEFAULT_TYPE`.
        veq (:obj:`str`, optional):
            Exact version that API object module file must match.

            Defaults to: "".
        vmin (:obj:`str`, optional):
            Minimum version that API object module file must match.

            Defaults to: "".
        vmax (:obj:`str`, optional):
            Maximum version that API object module file must match.

            Defaults to: "".
        vshrink (:obj:`bool`, optional):
            If True, shrink API object module file version down to match the
            same length as veq, vmax, and vmin.

            If False, do not shrink API object module file version at all.

            If int, shrink API object module file version, veq, vmax, and vmin
            down to this length.

            Defaults to: True.

    Raises:
        :exc:`exceptions.NoVersionFoundError`:
              If no API module files match api_type, veq, vmax, and vmin.

    Returns:
        :obj:`dict`

    """
    versions = get_versions(api_type=api_type)
    for version in versions:
        if utils.versions.version_check(
            version=version["ver_parts"], veq=veq, vmax=vmax, vmin=vmin, vshrink=vshrink
        ):
            return version

    error = [
        "",
        "Unable to find any {t!r} object modules where version is:",
        "  Equal to: {veq!r}",
        "  At least: {vmin!r}",
        "  At Most: {vmax!r}",
        "  File pattern matches: {p!r}",
        "",
        "Valid versions for type {t!r}:{vers}",
    ]
    vers = [v["ver_str"] for v in versions]
    error = "\n".join(error).format(
        t=api_type,
        p=PATTERN.format(api_type=api_type),
        veq=veq or "any",
        vmin=vmin or "any",
        vmax=vmax or "any",
        vers="\n  - " + "\n  - ".join(vers),
    )
    raise exceptions.NoVersionFoundError(error)


def load(api_type=DEFAULT_TYPE, veq="", vmax="", vmin="", vshrink=True):
    """Import an API module for version and type.

    Args:
        api_type (:obj:`str`, optional):
            Type of object module to load, must be one of :data:`API_TYPES`.

            Defaults to: :data:`DEFAULT_TYPE`.
        veq (:obj:`str`, optional):
            Exact version of object module of type to load.

            Defaults to: "".
        vmin (:obj:`str`, optional):
            Minimum version of object module of type to load.

            Defaults to: "".
        vmax (:obj:`str`, optional):
            Maximum version of object module of type to load.

            Defaults to: "".
        vshrink (:obj:`bool` or :obj:`int`, optional):
            If True, shrink API object module file version down to match the
            same length as veq, vmax, and vmin.

            If False, do not shrink API object module file version at all.

            If int, shrink API object module file version, veq, vmax, and vmin
            down to this length.

            Defaults to: True.

    Raises:
        :exc:`exceptions.ModuleError`:
             If api_type not in :data:`API_TYPES`.

    Returns:
        :obj:`ApiObjects`

    """
    api_type = api_type.lower()
    if api_type not in API_TYPES:
        error = "type {t!r} must be one of {a}"
        error = error.format(t=api_type, a=API_TYPES)
        raise exceptions.ModuleError(error)

    version = find_version(
        api_type=api_type, veq=veq, vmin=vmin, vmax=vmax, vshrink=vshrink
    )
    ret = ApiObjects(module_file=version["module_file"])
    return ret



# -*- coding: utf-8 -*-
"""Version information for tantrum."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

__version__ = "1.0.2"
""":obj:`str`: Version of package."""
__url__ = "https://github.com/lifehackjim/tantrum"
""":obj:`str`: URL of package."""
__author__ = "Jim Olsen"
""":obj:`str`: Auth of package."""
__title__ = "tantrum"
""":obj:`str`: Title of package."""
__project__ = "tantrum"
""":obj:`str`: Name of package."""
__author_email__ = "jim@lifehack.com"
""":obj:`str`: Author email of package."""
__description__ = "Tanium API workflow encapsulation"
""":obj:`str`: Description of package."""
__license__ = "MIT"
""":obj:`str`: License of package."""
__copyright__ = "Copyright Jim Olsen 2019"
""":obj:`str`: Copyright of package."""



# -*- coding: utf-8 -*-
"""Parent exceptions and warnings for the tantrum package."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


class TantrumError(Exception):
    """Parent exception for all :mod:`tantrum` errors."""

    pass


class TantrumWarning(Warning):
    """Parent warning for all :mod:`tantrum` warnings."""

    pass



# -*- coding: utf-8 -*-
"""Exceptions and warnings for :mod:`tantrum.adapters`."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .. import exceptions


class ModuleError(exceptions.TantrumError):
    """Parent of all exceptions for :mod:`tantrum.adapters`.

    Thrown by:
      * :func:`tantrum.adapters.load_type`
      * :func:`tantrum.adapters.load`

    """

    pass


class ModuleWarning(exceptions.TantrumWarning):
    """Parent of all warnings for :mod:`tantrum.adapters`.

    Thrown by:

    """

    pass


class InvalidTypeError(ModuleError):
    """Thrown when an object of an invalid type is supplied.

    Thrown by:
      * :func:`tantrum.adapters.Adapter.api_get_audit_logs`
      * :func:`tantrum.adapters.check_object_type`

    """

    pass


class EmptyAttributeError(ModuleError):
    """Thrown when an object is supplied that does not have required attributes set.

    Thrown by:
      * :func:`tantrum.adapters.check_object_attrs`

    """

    pass


class TypeMismatchError(ModuleError):
    """Thrown when an API module type does not match an adapters type.

    Thrown by:
      * :func:`tantrum.adapters.check_adapter_types`

    """

    pass


class SessionNotFoundWarning(ModuleWarning):
    """Thrown when a session XML tag can not be found in a response body.

    Thrown by:
      * :func:`tantrum.adapters.Soap.send`

    """

    pass



# -*- coding: utf-8 -*-
"""Exceptions and warnings for :mod:`tantrum.results`."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .. import exceptions
from .. import utils


class ModuleError(exceptions.TantrumError):
    """Parent of all exceptions for :mod:`tantrum.results`."""

    pass


class ModuleWarning(exceptions.TantrumWarning):
    """Parent of all warnings for :mod:`tantrum.results`."""

    pass


class ResponseError(ModuleError):
    """Thrown when a response has an general/uncategorized error."""

    def __init__(self, result, error):
        """Constructor.

        Args:
            result (:obj:`tantrum.results.Result`):
                Object exception was thrown from.
            error (:obj:`str`):
                Error string.
        """
        self.result = result
        """:obj:`tantrum.results.Result`: Object exception was thrown from."""
        self.error = error
        """:obj:`str`: Error string."""
        self.response_body = utils.tools.trim_txt(
            txt=result.response_body_str.strip(), limit=10000
        )
        """:obj:`str`: Trimmed response body."""

        error = [
            "",
            "-- From: {o}".format(o=result),
            "-- Request body:\n{o}".format(o=result.request_body_str),
            "-- Response body:\n{o}".format(o=self.response_body),
            "-- Error: {o}".format(o=error),
        ]
        self.error = "\n".join(error)
        """:obj:`str`: Error message that was thrown."""
        super(ResponseError, self).__init__(self.error)


class ObjectExistsError(ResponseError):
    """Thrown when a response tells us the object already exists."""

    pass


class ObjectNotFoundError(ResponseError):
    """Thrown when a response tells us the object already exists."""

    pass


class TextDeserializeError(ModuleError):
    """Thrown when deserializing a string into python object."""

    def __init__(self, result, text, src, exc):
        """Constructor.

        Args:
            result (:obj:`tantrum.results.Result`):
                Object exception was thrown from.
            text (:obj:`str`):
                Text that was being deserialized into python object.
            src (:obj:`str`):
                Description of where text came from.
            exc (:obj:`Exception`):
                Exception that occurred during deserialization.
        """
        self.result = result
        """:obj:`tantrum.results.Result`: Object exception was thrown from."""
        self.text = text
        """:obj:`str`: Text that was being deserialized into python object."""
        self.src = src
        """:obj:`str`: Description of where text came from."""
        self.exc = exc
        """:obj:`Exception`: Exception that occurred during deserialization."""

        error = [
            "",
            "-- From: {o}".format(o=result),
            "{src} text:".format(src=src),
            utils.tools.trim_txt(txt=text, limit=10000),
            "-- Unable to deserialize {src} text, error:".format(src=src),
            format(exc),
        ]
        self.error = "\n".join(error)
        """:obj:`str`: Error message that was thrown."""
        super(TextDeserializeError, self).__init__(self.error)


class ApiWrongRequestType(ModuleError):
    """Thrown when get result data/object API obj but request is not of that type."""

    def __init__(self, result, error):
        """Constructor.

        Args:
            result (:obj:`tantrum.results.Result`):
                Object exception was thrown from.
        """
        self.result = result
        """:obj:`tantrum.results.Result`: Object exception was thrown from."""

        error = ["", "-- From: {o}".format(o=result), "-- Error: {o}".format(o=error)]
        self.error = "\n".join(error)
        """:obj:`str`: Error message that was thrown."""
        super(ApiWrongRequestType, self).__init__(self.error)


class DictionaryPathError(ModuleError):
    """Thrown when a response has an error when deserializing."""

    def __init__(self, result, obj, path, src, exc):
        """Constructor.

        Args:
            result (:obj:`tantrum.results.Result`):
                Object exception was thrown from.
            obj (:obj:`dict`):
                Dictionary to traverse using path.
            path (:obj:`str`):
                Nested dictionary keys seperated by / to traverse in obj.
            src (:obj:`str`):
                Where obj came from, used in error text.
            exc (:obj:`Exception`):
                Original exception thrown.
        """
        self.result = result
        """:obj:`tantrum.results.Result`: Object exception was thrown from."""
        self.obj = obj
        """:obj:`dict`: Dictionary to traverse using path."""
        self.path = path
        """:obj:`str`: Nested dictionary keys seperated by / to traverse in obj."""
        self.src = src
        """:obj:`str`: Where obj came from."""
        self.exc = exc
        """:obj:`Exception`: Original exception thrown."""

        error = [
            "",
            "Unable to find path '{path}' in {src}:".format(path=path, src=src),
            format(exc),
        ]
        self.error = "\n".join(error)
        """:obj:`str`: Error message that was thrown."""
        super(DictionaryPathError, self).__init__(self.error)



# -*- coding: utf-8 -*-
"""Exceptions and warnings for :mod:`tantrum.auth_methods`."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .. import exceptions


class ModuleError(exceptions.TantrumError):
    """Parent of all exceptions for :mod:`tantrum.auth_methods`."""

    pass


class ModuleWarning(exceptions.TantrumWarning):
    """Parent of all warnings for :mod:`tantrum.auth_methods`."""

    pass


class MethodError(ModuleError):
    """Parent of all :obj:`tantrum.auth_methods.AuthMethod` exceptions."""

    msg = "Authentication Method Error!"
    """:obj:`str`: Error message to use in exception."""

    def __init__(self, auth_method, info=None, response=None):
        """Constructor.

        Args:
            auth_method (:obj:`tantrum.auth_methods.AuthMethod`):
                Object where exception was thrown from.
            info (:obj:`str`, optional):
                Additional error info message to show.

                Defaults to: None.
            response (:obj:`requests.Response`, optional):
                Response associated with this exception, if any.

                Defaults to: None.
        """
        self.auth_method = auth_method
        self.response = response

        msgs = ["AuthMethod: {a}".format(a=auth_method)]

        if response is not None:
            msgs += [
                "Response body: {r}".format(r=response.text),
                "Request URL: {r!r}".format(r=response.url),
                "Request method: {r!r}".format(r=response.request.method),
                "Response status code: {r!r}".format(r=response.status_code),
            ]

        msgs.append("Error: {m}".format(m=self.msg))
        error = "\n\t" + "\n\t".join(msgs)
        self.error = error
        """:obj:`str`: Error message that was thrown."""
        super(MethodError, self).__init__(error)


class InvalidToken(MethodError):
    """Thrown when validate_token fails."""

    msg = "Supplied token is invalid!"
    """:obj:`str`: Error message to use in exception."""


class NotLoggedInError(MethodError):
    """Thrown when logout called before login."""

    msg = "Not logged in, unable to logout/validate token. Must login first!"
    """:obj:`str`: Error message to use in exception."""


class LoginError(MethodError):
    """Thrown when login response has any status_code other than 200."""

    msg = "Login failed. Response status code is not 200!"
    """:obj:`str`: Error message to use in exception."""


class LogoutError(MethodError):
    """Thrown when logout fails."""

    msg = "Logout failed. Verify the supplied credentials!"
    """:obj:`str`: Error message to use in exception."""



# -*- coding: utf-8 -*-
"""Authentication methods for the Tanium API."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import abc
import six
import re

from datetime import datetime, timedelta

from . import exceptions
from .. import utils


@six.add_metaclass(abc.ABCMeta)
class AuthMethod(object):
    """Abstract base class for all AuthMethods."""

    @abc.abstractproperty
    def http_client(self):
        """Get the HTTP Client for this object.

        Returns:
            :obj:`tantrum.http_client.HttpClient`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractproperty
    def token(self):
        """Get or set a token.

        Notes:
            If no token received, :meth:`login` to get one.

            If token expired, :meth:`login` to get a new one.

            If token has been received, :meth:`validate` to validate token.

            If token fails validation, :meth:`login` to get a new one.

        Returns:
            :obj:`str`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractproperty
    def token_headers(self):
        """Get dict with :attr:`token` for use in headers.

        Returns:
            :obj:`dict`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractproperty
    def uid(self):
        """Get user ID for token if a token has been received.

        Returns:
            :obj:`str`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractproperty
    def logged_in(self):
        """Check if a token has been received.

        Returns:
            :obj:`bool`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def login(self, **kwargs):
        """Send a login request to receive a token.

        Raises:
            :exc:`exceptions.LoginError`:
                If status code in response is not 200.

        Returns:
            :obj:`str`

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def logout(self, **kwargs):
        """Send a logout request to revoke a token.

        Raises:
            :exc:`exceptions.NotLoggedInError`:
                If :attr:`logged_in` is False.
            :exc:`exceptions.LogoutError`:
                If status code in response is not 200.

        Returns:
            None

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def logout_all(self, **kwargs):
        """Send a logout request to revoke all tokens associated with this token.

        Raises:
            :exc:`exceptions.NotLoggedInError`:
                If :attr:`logged_in` is False.
            :exc:`exceptions.LogoutError`:
                If status code in response is not 200.

        Returns:
            None

        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def validate(self, **kwargs):
        """Send a validate request to check that token is still valid.

        Raises:
            :exc:`exceptions.NotLoggedInError`:
                If :attr:`logged_in` is False.
            :exc:`exceptions.InvalidToken`:
                If status code in response is not 200.

        Returns:
            :obj:`str`

        """
        raise NotImplementedError  # pragma: no cover


class CommonMixin(object):
    """Shared methods common amongst all :class:`AuthMethod`."""

    def __init__(
        self,
        http_client,
        login_timeout=5,
        logout_timeout=5,
        expires_after=295,
        lvl="info",
    ):
        """Constructor.

        Args:
            http_client (:obj:`tantrum.http_client.HttpClient`):
                Object for sending HTTP requests.
            login_timeout (:obj:`int`, optional):
                Timeout for login and validate responses in seconds.

                Defaults to: 5.
            logout_timeout (:obj:`int`, optional):
                Timeout for logout responses in seconds.

                Defaults to: 5.
            expires_after (:obj:`int`, optional):
                Life of received tokens in seconds.

                Defaults to: 295.
            lvl (:obj:`str`, optional):
                Logging level for this object.

                Defaults to: "info".

        """
        self.log = utils.logs.get_obj_log(obj=self, lvl=lvl)
        """:obj:`logging.Logger`: Log for this object."""

        self.revalidate_after = 5
        """:obj:`int`: Revalidate token if :attr:`last_used` is higher than this."""

        self._http_client = http_client
        self._login_timeout = login_timeout
        self._logout_timeout = logout_timeout
        self._expires_after = expires_after
        self._last_used = None
        self._token = None
        self.token = None

    def __str__(self):
        """Show object info.

        Returns:
            :obj:`str`

        """
        bits = [
            "logged_in={}".format(self.logged_in),
            "last_used_secs={}".format(self.last_used_secs),
            "expiry_dt={}".format(self.expiry_dt),
        ]
        bits = "({})".format(", ".join(bits))
        cls = "{c.__module__}.{c.__name__}".format(c=self.__class__)
        return "{cls}{bits}".format(cls=cls, bits=bits)

    def __repr__(self):
        """Show object info.

        Returns:
            :obj:`str`

        """
        return self.__str__()

    @property
    def http_client(self):
        """Get the HTTP Client for this object.

        Returns:
            :obj:`tantrum.http_client.HttpClient`

        """
        return self._http_client

    @property
    def token(self):
        """Get a token.

        Notes:
            If :attr:`logged_in` is False, :meth:`login` to get one.

            If :attr:`expired` is True, :meth:`login` to get a new one.

            If :attr:`logged_in` is True, and :attr:`last_used_secs` is older than
            :attr:`revalidate_after`, :meth:`validate` to re-validate token.

            If :meth:`validate` fails validation, :meth:`login` to get a new one.

        Returns:
            :obj:`str`

        """
        if not self.logged_in:
            self.login(cause="Initial login")
        elif self.expired:
            self.login(cause="Re-login due to expired token")
        else:
            if self.last_used_secs >= self.revalidate_after:
                try:
                    self.validate(cause="Validate existing token")
                except exceptions.InvalidToken:
                    m = "Token for User ID {uid!r} no longer valid, getting new one"
                    m = m.format(uid=self.uid)
                    self.log.debug(m)
                    self.login(cause="Re-login due to invalid token")
        return self._token

    @token.setter
    def token(self, value):
        """Set the token to value.

        Args:
            value (:obj:`str`):
                Value to set _token to

        """
        self._token = value
        self._last_used = None if value is None else datetime.utcnow()

    @property
    def token_headers(self):
        """Get dict with :attr:`token` for use in headers.

        Returns:
            :obj:`dict`

        """
        return {"session": self.token}

    @property
    def uid(self):
        """Get user ID for token if a token has been received.

        Returns:
            :obj:`str`

        """
        return int((self._token or self.token).split("-")[0])

    @property
    def logged_in(self):
        """Check if a token has been received.

        Returns:
            :obj:`bool`

        """
        return True if self._token else False

    @property
    def expiry_dt(self):
        delta = timedelta(seconds=self._expires_after)
        return None if not self._last_used else self._last_used + delta

    @property
    def expired(self):
        """Check if token has expired.

        Returns:
            :obj:`bool`

        """
        return True if not self._last_used else datetime.utcnow() >= self.expiry_dt

    @property
    def last_used_secs(self):
        """Get the number of seconds since token was issued or last used.

        Returns:
            :obj:`int`

        """
        return (
            0 if not self._last_used else (datetime.utcnow() - self._last_used).seconds
        )

    def login(self, **kwargs):
        """Send a login request to receive a token.

        Args:
            **kwargs:
                cause (:obj:`str`):
                    String to explain purpose of request.

                    Defaults to: "Get new token".

        Raises:
            :exc:`exceptions.LoginError`:
                If status code in response is not 200.

        Returns:
            :obj:`str`

        """
        r = self.http_client(
            path="/auth",
            method="post",
            headers={k: v for k, v in self._headers.items() if v},
            b64_headers=self._b64_headers,
            timeout=self._login_timeout,
            cause=kwargs.pop("cause", "Get new token"),
        )

        if r.status_code != 200:
            raise exceptions.LoginError(auth_method=self, response=r)

        self.token = r.text
        m = "Token received for User ID {uid!r} from {url!r}"
        m = m.format(uid=self.uid, url=r.url)
        self.log.debug(m)
        return r.text

    def logout(self, **kwargs):
        """Send a logout request to revoke a token.

        Args:
            **kwargs:
                cause (:obj:`str`):
                    String to explain purpose of request.

                    Defaults to: "Revoke token".
                headers (:obj:`dict`):
                    Headers to send in request.

                    Defaults to: {}.

        Raises:
            :exc:`exceptions.NotLoggedInError`:
                If :attr:`logged_in` is False.
            :exc:`exceptions.LogoutError`:
                If status code in response is not 200.

        Returns:
            None

        """
        if not self.logged_in:
            raise exceptions.NotLoggedInError(auth_method=self)

        r = self.http_client(
            path="/auth",
            method="post",
            headers={"session": self._token, "logout": "0"},
            b64_headers=self._b64_headers,
            timeout=self._logout_timeout,
            cause=kwargs.pop("cause", "Revoke token"),
        )

        if r.status_code != 200:
            raise exceptions.LogoutError(auth_method=self, response=r)

        m = "Token revoked for User ID {uid!r} from {url!r}"
        m = m.format(uid=self.uid, url=r.url)
        self.log.debug(m)
        self.token = None
        return None

    def logout_all(self, **kwargs):
        """Send a logout request to revoke all tokens associated with this token.

        Args:
            **kwargs:
                cause (:obj:`str`):
                    String to explain purpose of request.

                    Defaults to: "Revoke all tokens".
                headers (:obj:`dict`):
                    Headers to send in request.

                    Defaults to: {}.

        Raises:
            :exc:`exceptions.NotLoggedInError`:
                If :attr:`logged_in` is False.
            :exc:`exceptions.LogoutError`:
                If status code in response is not 200.

        Returns:
            None

        """
        if not self.logged_in:
            raise exceptions.NotLoggedInError(auth_method=self)

        r = self.http_client(
            path="/auth",
            method="post",
            headers={"session": self._token, "logout": "1"},
            b64_headers=self._b64_headers,
            timeout=self._logout_timeout,
            cause=kwargs.pop("cause", "Revoke all tokens"),
        )

        if r.status_code != 200:
            raise exceptions.LogoutError(auth_method=self, response=r)

        m = "All tokens revoked for User ID {uid!r} from {url!r}"
        m = m.format(uid=self.uid, url=r.url)
        self.log.debug(m)
        self.token = None
        return None

    def validate(self, **kwargs):
        """Send a validate request to check that token is still valid.

        Args:
            **kwargs:
                cause (:obj:`str`):
                    String to explain purpose of request.

                    Defaults to: "Validate token".
                headers (:obj:`dict`):
                    Headers to send in request.

                    Defaults to: {}.

        Raises:
            :exc:`exceptions.NotLoggedInError`:
                If :attr:`logged_in` is False.
            :exc:`exceptions.InvalidToken`:
                If status code in response is not 200.

        Returns:
            :obj:`str`

        """
        if not self.logged_in:
            raise exceptions.NotLoggedInError(auth_method=self)

        r = self.http_client(
            path="/auth",
            method="post",
            headers={"session": self._token},
            timeout=self._login_timeout,
            cause=kwargs.pop("cause", "Validate token"),
        )

        if r.status_code != 200:
            raise exceptions.InvalidToken(auth_method=self, response=r)

        m = "Token validated for User ID {uid!r} from {url!r}"
        m = m.format(uid=self.uid, url=r.url)
        self.log.debug(m)
        self.token = r.text
        return r.text


class Credentials(CommonMixin, AuthMethod):
    """Method that uses credentials to interact with the '/auth' API."""

    def __init__(
        self,
        http_client,
        username,
        password,
        domain="",
        secondary="",
        login_timeout=5,
        logout_timeout=5,
        expires_after=295,
        lvl="info",
    ):
        """Constructor.

        Args:
            http_client (:obj:`tantrum.http_client.HttpClient`):
                HTTP client.
            username (:obj:`str`):
                Header to pass to /auth API.
            password (:obj:`str`):
                Header to pass to /auth API.
            domain (:obj:`str`, optional):
                Header to pass to /auth API.

                Defaults to: ""
            secondary (:obj:`str`, optional):
                Header to pass to /auth API.

                Defaults to: ""
            login_timeout (:obj:`int`, optional):
                Timeout for login and validate responses in seconds.

                Defaults to: 5.
            logout_timeout (:obj:`int`, optional):
                Timeout for logout responses in seconds.

                Defaults to: 5.
            expires_after (:obj:`int`, optional):
                Life of received tokens in seconds.

                Defaults to: 295.
            lvl (:obj:`str`, optional):
                Logging level for this object.

                Defaults to: "info".

        """
        self._headers = {
            "username": username,
            "password": password,
            "domain": domain,
            "secondary": secondary,
        }
        self._b64_headers = ["username", "password"]
        super(Credentials, self).__init__(
            http_client=http_client,
            login_timeout=login_timeout,
            logout_timeout=logout_timeout,
            expires_after=expires_after,
            lvl=lvl,
        )


class SessionId(CommonMixin, AuthMethod):
    """Method that uses session id to interact with the '/auth' API."""

    def __init__(
        self,
        http_client,
        session,
        login_timeout=5,
        logout_timeout=5,
        expires_after=295,
        lvl="info",
    ):
        """Constructor.

        Args:
            http_client (:obj:`tantrum.http_client.HttpClient`):
                HTTP client.
            session (:obj:`str`):
                Header to pass to /auth API.
            login_timeout (:obj:`int`, optional):
                Timeout for login and validate responses in seconds.

                Defaults to: 5.
            logout_timeout (:obj:`int`, optional):
                Timeout for logout responses in seconds.

                Defaults to: 5.
            expires_after (:obj:`int`, optional):
                Life of received tokens in seconds.

                Defaults to: 295.
            lvl (:obj:`str`, optional):
                Logging level for this object.

                Defaults to: "info".

        """
        self._headers = {"session": session}
        self._b64_headers = []
        super(SessionId, self).__init__(
            http_client=http_client,
            login_timeout=login_timeout,
            logout_timeout=logout_timeout,
            expires_after=expires_after,
            lvl=lvl,
        )
        self.token = session

    @property
    def token(self):
        """Get a token.

        Notes:
            If :attr:`CommonMixin.logged_in` is False, can not login to get a
            new one with just a session.

            If :attr:`CommonMixin.expired` is True, can not login to get
            a new one with just a session.

            If :attr:`CommonMixin.logged_in` is True, and
            :attr:`CommonMixin.last_used_secs` is older than
            :attr:`CommonMixin.revalidate_after`,
            :meth:`CommonMixin.validate` to re-validate token.

            If :meth:`CommonMixin.validate` fails validation, can not login
            to get a new one with just a session.

        Returns:
            :obj:`str`

        """
        if self.last_used_secs is None or self.last_used >= self.revalidate_after:
            self.validate(cause="Validate existing token")
        return self._token

    @token.setter
    def token(self, value):
        """Set the token to value and :attr:`CommonMixin.last_used` to now.

        Args:
            value (:obj:`str`):
                Value to set _token to

        """
        self._token = value
        if value is None:
            self._last_used = None
            self._last_validated = None
        else:
            self._last_used = datetime.utcnow()
            self._last_validated = datetime.utcnow()

    def login(self, **kwargs):
        """Send a login request to receive a token.

        Raises:
            :exc:`exceptions.LoginError`:
                If status code in response is not 200.

        Returns:
            :obj:`str`

        """
        error = "Unable to perform a login with a session ID, can only validate"
        raise NotImplementedError(error)


def validate_token(token):
    """Validate that a token is properly formed.

    Args:
        token (:obj:`str`):
            Token to validate

    Returns:
        :obj:`bool`

    """
    return re.match(r"(\d+-\d+-)\w+", token) is not None



# -*- coding: utf-8 -*-
"""Exceptions and warnings for :mod:`tantrum.api_models`."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import pprint

from .. import exceptions


class ModuleError(exceptions.TantrumError):
    """Parent of all exceptions for :mod:`tantrum.api_models`."""

    pass


class ModuleWarning(exceptions.TantrumWarning):
    """Parent of all warnings for :mod:`tantrum.api_models`."""

    pass


class AttrUndefinedWarning(ModuleWarning):
    """Warning handler for a simple value being set on an undefined attribute."""

    def __init__(self, obj, attr, value):
        """Constructor.

        Args:
            obj (:obj:`tantrum.api_models.ApiModel`):
                Object that warning was thrown from.
            attr (:obj:`str`):
                Simple attribute that was adhoc added.
            value (:obj:`int` or :obj:`str` or :obj:`bool`):
                Value of attr.

        """
        self.obj = obj
        """:obj:`tantrum.api_models.ApiModel`: Object that warning was thrown from."""
        self.attr = attr
        """:obj:`str`: Simple attribute that was adhoc added."""
        self.value = value
        """:obj:`int` or :obj:`str` or :obj:`bool`: Value of attr."""

        error = [
            "Adhoc added undefined simple attribute: "
            "{attr!r} of type {valtype!r} to {cls!r}"
        ]
        error = "\n".join(error)
        self.error = error.format(
            cls=obj.__class__.__name__, attr=attr, valtype=type(value).__name__
        )
        """:obj:`str`: Warning message that was thrown."""
        super(AttrUndefinedWarning, self).__init__(self.error)


class AttrUndefinedError(ModuleError):
    """Exception handler for a complex value being set on an undefined attribute."""

    def __init__(self, obj, attr, value):
        """Constructor.

        Args:
            obj (:obj:`tantrum.api_models.ApiModel`):
                Object that exception was thrown from.
            attr (:obj:`str`):
                Complex attribute that was supplied.
            value (:obj:`dict` or :obj:`list`):
                Value of attr.

        """
        self.obj = obj
        """:obj:`tantrum.api_models.ApiModel`: Object that exception was thrown from."""
        self.attr = attr
        """:obj:`str`: Complex attribute that was supplied."""
        self.value = value
        """:obj:`dict` or :obj:`list`: Value of attr."""

        error = [
            "",
            "{def_attrs}",
            "Received attribute {attr!r} of type {valtype!r} on {cls!r}",
            "But attribute is not defined as a complex attribute.",
        ]
        error = "\n".join(error)
        self.error = error.format(
            cls=obj.__class__,
            attr=attr,
            valtype=type(value).__name__,
            def_attrs="\n".join(obj.api_attrs_desc()),
        )
        """:obj:`str`: Error message that was thrown."""
        super(AttrUndefinedError, self).__init__(self.error)


class AttrTypeError(ModuleError):
    """Exception handler for type mismatch on a defined attribute."""

    def __init__(self, obj, attr, value, be_type):
        """Constructor.

        Args:
            obj (:obj:`tantrum.api_models.ApiModel`):
                Object that exception was thrown from.
            attr (:obj:`str`):
                Attribute that had a type mismatch.
            value (:obj:`object`):
                Value of attr.
            be_type (:class:`object`):
                Type that value should be, but is not.

        """
        self.obj = obj
        """:obj:`tantrum.api_models.ApiModel`: Object that exception was thrown from."""
        self.attr = attr
        """:obj:`str`: Attribute that had a type mismatch."""
        self.value = value
        """:obj:`object`: Value of attr."""
        self.be_type = be_type
        """:class:`object`: Type that value should be, but is not."""

        error = [
            "",
            "Received attribute {attr!r} of type {valtype!r} on {cls!r}",
            "Type for attribute {attr!r} must be: {be_type!r}",
        ]
        error = "\n".join(error)
        self.error = error.format(
            cls=obj.__class__.__name__,
            attr=attr,
            be_type=be_type,
            valtype=type(value).__name__,
        )
        """:obj:`str`: Error message that was thrown."""
        super(AttrTypeError, self).__init__(self.error)


class ListTypeError(ModuleError):
    """Exception handler for type mismatch on a list action."""

    def __init__(self, obj, attr, value, op):
        """Constructor.

        Args:
            obj (:obj:`tantrum.api_models.ApiList`):
                Object that exception was thrown from.
            attr (:obj:`str`):
                Attribute that had a type mismatch.
            value (:obj:`object`):
                Value of attr.
            op (:obj:`str`):
                Operation being performed with type mismatch.

        """
        self.obj = obj
        """:obj:`tantrum.api_models.ApiList`: Object that exception was thrown from."""
        self.attr = attr
        """:obj:`str`: Attribute that had a type mismatch."""
        self.value = value
        """:obj:`object`: Value of attr."""
        self.op = op
        """:obj:`str`: Operation being performed with type mismatch."""

        atmpl = "on attribute {a!r} ".format(a=attr) if attr else ""

        error = [
            "",
            "While performing operation {op!r} on {cls!r} on attribute {attr}",
            "Invalid type of list {valtype!r} with value:",
            "{value}",
            "Must be a list, tuple, or an instance of {cls!r}",
        ]
        error = "\n".join(error)
        self.error = error.format(
            cls=obj.__class__.__name__,
            attr=atmpl,
            value=pprint.pformat(value),
            valtype=type(value).__name__,
            op=op,
        )
        """:obj:`str`: Error message that was thrown."""
        super(ListTypeError, self).__init__(self.error)


class ListItemTypeError(ModuleError):
    """Exception handler for type mismatch when adding an item to a list."""

    def __init__(self, obj, item, items, attr, op, be_type):
        """Constructor.

        Args:
            obj (:obj:`tantrum.api_models.ApiList`):
                Object that exception was thrown from.
            item (:obj:`object`):
                Item being added in items that is the wrong type.
            items (:obj:`list`):
                All items being added to list.
            attr (:obj:`str`):
                Attribute that had a type mismatch.
            op (:obj:`str`):
                Operation being performed with type mismatch.
            be_type (:class:`object`):
                Type that item should be, but is not.

        """
        self.obj = obj
        """:obj:`tantrum.api_models.ApiList`: Object that exception was thrown from."""
        self.item = item
        """:obj:`object`: Item being added in items that is the wrong type."""
        self.items = items
        """:obj:`list`: Full list of items containing item passed to op."""
        self.attr = attr
        """:obj:`str`: Attribute that had a type mismatch."""
        self.op = op
        """:obj:`str`: Operation being performed with type mismatch."""
        self.be_type = be_type
        """:class:`object`: Type that item should be, but is not."""

        error = [
            "",
            "While perform operation {op!r} on {cls!r}{attr}",
            "On item in a {lsttype} with {lstcnt} items",
            "Invalid type on item: {valtype!r} with value: {item!r}",
            "Item must be of type: {be_type}",
        ]
        error = "\n".join(error)
        self.error = error.format(
            cls=obj.__class__.__name__,
            attr=" on attribute {a!r} ".format(a=attr) if attr else "",
            item=item,
            lsttype=type(items),
            lstcnt=0 if items is None else len(items),
            be_type=be_type,
            op=op,
            valtype=type(item).__name__,
        )
        """:obj:`str`: Error message that was thrown."""
        super(ListItemTypeError, self).__init__(self.error)


class GetSingleItemError(ModuleError):
    """Exception handler when no items found."""

    def __init__(self, obj, attr, value, regex_value, items):
        """Constructor.

        Args:
            obj (:obj:`tantrum.api_models.ApiList`):
                Object that exception was thrown from.
            attr (:obj:`str`):
                Attribute to search for value of in items.
            value (:obj:`object`):
                Value of attr that was being used to match against items.
            regex_value (:obj:`bool`):
                If regex matching was being used for value.
            items (:obj:`list`):
                All items in list object.

        """
        self.obj = obj
        """:obj:`tantrum.api_models.ApiList`: Object that exception was thrown from."""
        self.attr = attr
        """:obj:`str`: Attribute used to search for value of in items."""
        self.value = value
        """:obj:`object`: Value of attr that was being used to match against items."""
        self.regex_value = regex_value
        """:obj:`bool`: If regex matching was being used for value."""
        self.items = items
        """:obj:`list`: All items found in list object."""

        itemvals = [format(getattr(x, attr, None)) for x in obj]
        itemvals = "\n  ".join(itemvals or ["Empty list!"])

        error = [
            "",
            "All values for attribute {attr!r} type {attrtype} in {obj}:",
            "  {itemvals}",
            "",
            "Found {cnt} items while searching for:",
            " 1 item where {attr!r} {op} value {value!r} of type {vt}:",
        ]

        error = "\n".join(error)
        self.error = error.format(
            attr=attr,
            attrtype=obj.API_ITEM_CLS.api_attrs()[attr],
            obj=obj,
            itemvals=itemvals,
            value=value,
            vt=type(value),
            op="regex matches" if regex_value else "equals",
            cnt=len(items),
        )
        """:obj:`str`: Error message that was thrown."""
        super(GetSingleItemError, self).__init__(self.error)
